from __future__ import annotations

import inspect
import logging
from functools import wraps
from typing import Any, Callable, Awaitable, Tuple, TypeVar, ParamSpec, cast

PARAMS_SPEC = ParamSpec("PARAMS_SPEC")
RESULT_SPEC = TypeVar("RESULT_SPEC")

_logger = logging.getLogger(__name__)


def raise_own_exception_if_exception(
        *,
        catching_exceptions: type[BaseException] | Tuple[type[BaseException], ...] | None = Exception,
        except_catching_exceptions: type[BaseException] | Tuple[type[BaseException], ...] | None = None,
        own_exception: type[Exception],
        kwargs_in_own_exception: dict[str, Any] | None = None,
        forward_kwargs_in_own_exception: dict[str, Any] | None = None,
) -> (
        Callable[[Callable[PARAMS_SPEC, RESULT_SPEC] | Callable[PARAMS_SPEC, Awaitable[RESULT_SPEC]]],
        Callable[PARAMS_SPEC, RESULT_SPEC] | Callable[PARAMS_SPEC, Awaitable[RESULT_SPEC]]]
):
    """
    Ловит любые Exception и оборачивает в own_exception(**kwargs_).
    Исключения из except_exceptions пропускает как есть.
    Работает и для sync, и для async функций.
    """

    # normalize catching_exceptions -> tuple[type, ...]
    if catching_exceptions is None:
        catching_exceptions = (Exception,)
    elif isinstance(catching_exceptions, tuple):
        catching_exceptions = catching_exceptions
    else:
        catching_exceptions = (catching_exceptions,)

    # Нормализуем except_exceptions к кортежу типов
    if except_catching_exceptions is None:
        except_catching_exceptions = ()
    elif isinstance(except_catching_exceptions, tuple):
        except_catching_exceptions = except_catching_exceptions
    elif except_catching_exceptions:
        except_catching_exceptions = (except_catching_exceptions,)
    else:
        except_catching_exceptions = ()

    if kwargs_in_own_exception is not None:
        kwargs_in_own_exception = dict(kwargs_in_own_exception or {})
        kwargs_in_own_exception["catching_exceptions"] = catching_exceptions
        kwargs_in_own_exception["except_catching_exceptions"] = except_catching_exceptions
        kwargs_in_own_exception["own_exception"] = own_exception

    # Если явно передали пустой набор для ловли — возвращаем функцию как есть
    if not catching_exceptions:
        def _passthrough_decorator(func):
            return func

        return _passthrough_decorator

    def decorator(func: Callable[PARAMS_SPEC, RESULT_SPEC] | Callable[PARAMS_SPEC, Awaitable[RESULT_SPEC]]):
        if inspect.iscoroutinefunction(func):
            @wraps(func)
            async def async_wrapper(*args: PARAMS_SPEC.args, **kwargs: PARAMS_SPEC.kwargs) -> RESULT_SPEC:
                try:
                    return await cast(Callable[PARAMS_SPEC, Awaitable[RESULT_SPEC]], func)(*args, **kwargs)
                except catching_exceptions as caught_exception:  # ловим ТОЛЬКО нужные типы
                    if except_catching_exceptions and isinstance(caught_exception, except_catching_exceptions):
                        raise  # пропускаем как есть
                    if kwargs_in_own_exception is not None:
                        copied_kwargs_in_own_exception = kwargs_in_own_exception.copy()
                        copied_kwargs_in_own_exception["caught_exception"] = caught_exception
                        copied_kwargs_in_own_exception["caught_exception_type"] = str(type(caught_exception))
                        copied_kwargs_in_own_exception["caught_exception_str"] = str(caught_exception)
                    try:
                        _kwargs = {}
                        if kwargs_in_own_exception is not None:
                            _kwargs.update({"kwargs_in_own_exception": copied_kwargs_in_own_exception})
                        if forward_kwargs_in_own_exception is not None:
                            _kwargs.update(forward_kwargs_in_own_exception)
                        raise own_exception(**_kwargs) from caught_exception
                    except TypeError as exception:
                        _logger.warning(exception)
                        raise own_exception() from caught_exception

            return cast(Callable[PARAMS_SPEC, Awaitable[RESULT_SPEC]], async_wrapper)

        @wraps(func)
        def wrapper(*args: PARAMS_SPEC.args, **kwargs: PARAMS_SPEC.kwargs) -> RESULT_SPEC:
            try:
                return cast(Callable[PARAMS_SPEC, RESULT_SPEC], func)(*args, **kwargs)
            except catching_exceptions as caught_exception:
                if except_catching_exceptions and isinstance(caught_exception, except_catching_exceptions):
                    raise
                if kwargs_in_own_exception is not None:
                    copied_kwargs_in_own_exception = kwargs_in_own_exception.copy()
                    copied_kwargs_in_own_exception["caught_exception"] = caught_exception
                    copied_kwargs_in_own_exception["caught_exception_type"] = str(type(caught_exception))
                    copied_kwargs_in_own_exception["caught_exception_str"] = str(caught_exception)
                try:
                    _kwargs = {}
                    if kwargs_in_own_exception is not None:
                        _kwargs.update({"kwargs_in_own_exception": copied_kwargs_in_own_exception})
                    if forward_kwargs_in_own_exception is not None:
                        _kwargs.update(forward_kwargs_in_own_exception)
                    raise own_exception(**_kwargs) from caught_exception
                except TypeError as exception:
                    _logger.warning(exception)
                    raise own_exception() from caught_exception

        return cast(Callable[PARAMS_SPEC, RESULT_SPEC], wrapper)

    return decorator



# arpakit

from typing import Any

import pandas


def list_of_dicts_to_xlsx(
        list_of_dicts: list[dict[str, Any]],
        out_filepath: str = "out.xlsx",
        sheet_name: str = "Sheet 1"
) -> str:
    data_frame = pandas.DataFrame(list_of_dicts)

    writer = pandas.ExcelWriter(out_filepath, engine="xlsxwriter")
    data_frame.to_excel(writer, index=False, sheet_name=sheet_name)

    for i, col in enumerate(data_frame.columns):
        width = max(data_frame[col].apply(lambda x: len(str(x))).max(), len(col))
        writer.sheets[sheet_name].set_column(i, i, width * 1.1)

    writer._save()

    return out_filepath


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

import ipaddress


def is_ipv4_address(value: str) -> bool:
    try:
        ipaddress.IPv4Address(value)
    except ValueError:
        return False
    return True


def raise_if_not_ipv4_address(value: str):
    if not is_ipv4_address(value):
        raise ValueError(f"not is_ipv4_address({value})")


def is_ipv6_address(value: str) -> bool:
    try:
        ipaddress.IPv6Address(value)
    except ValueError:
        return False
    return True


def raise_if_not_ipv6_address(value: str):
    if not is_ipv6_address(value):
        raise ValueError(f"not is_ipv6_address({value})")


def is_ipv4_interface(value: str) -> bool:
    try:
        ipaddress.IPv4Interface(value)
    except ValueError:
        return False
    return True


def raise_if_not_ipv4_interface(value: str):
    if not is_ipv4_interface(value):
        raise ValueError(f"not is_ipv4_interface({value})")


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

import json
from typing import Any

from arpakitlib.ar_enumeration_util import Enumeration


class NeedTypes(Enumeration):
    str_ = "str"
    int_ = "int"
    bool_ = "bool"
    float_ = "float"
    list_of_int = "list_of_int"
    list_of_str = "list_of_str"
    list_of_float = "list_of_float"
    json = "json"


def parse_need_type(value: Any, need_type: str, allow_none: bool = False) -> Any:
    if allow_none and value is None:
        return None

    if not allow_none and value is None:
        raise ValueError("value is None")

    NeedTypes.parse_and_validate_values(need_type)

    if need_type == NeedTypes.str_:
        res = value
    elif need_type == NeedTypes.int_:
        res = int(value)
    elif need_type == NeedTypes.bool_:
        if value.lower() in ["true", "1"]:
            res = True
        elif value.lower() in ["false", "0"]:
            res = False
        else:
            raise ValueError(f"value {value} is not bool type")
    elif need_type == NeedTypes.float_:
        res = float(value)
    elif need_type == NeedTypes.list_of_int:
        res = value.removeprefix("[").removesuffix("]")
        res = [int(num.strip()) for num in res.split(",")]
    elif need_type == NeedTypes.list_of_str:
        res = value.removeprefix("[").removesuffix("]").strip()
        if not res:
            res = []
        else:
            res = [num.strip() for num in res.split(",")]
    elif need_type == NeedTypes.list_of_float:
        res = value.removeprefix("[").removesuffix("]")
        res = [float(num.strip()) for num in res.split(",")]
    elif need_type == NeedTypes.json:
        res = json.loads(value)
    else:
        raise ValueError(f"bad need_type {need_type}")

    return res


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

import json
from typing import Any, Iterable

import orjson


def transfer_json_str_to_data(
        json_str: str, fast: bool = False
) -> dict[Any, Any] | list[Any] | None:
    if not isinstance(json_str, str):
        raise ValueError("not isinstance(json_str, str)")
    if fast:
        return orjson.loads(json_str)
    else:
        return json.loads(json_str)


def transfer_data_to_json_str(
        data: dict[Any, Any] | list[Any] | None, beautify: bool = True, fast: bool = False
) -> str:
    if not isinstance(data, dict) and not isinstance(data, list) and data is not None:
        raise ValueError("not isinstance(data, dict) and not isinstance(data, list) and data is not None")
    if fast:
        return orjson.dumps(data).decode()
    else:
        if beautify:
            return json.dumps(data, ensure_ascii=False, indent=2, default=str)
        else:
            return json.dumps(data, ensure_ascii=False, default=str)


def transfer_data_to_json_str_to_data(
        data: dict[Any, Any] | list[Any] | None, fast: bool = False
) -> dict[Any, Any] | list[Any] | None:
    return transfer_json_str_to_data(transfer_data_to_json_str(data=data, fast=fast), fast=fast)


def transfer_json_str_to_data_to_json_str(
        json_str: str, beautify: bool = True, fast: bool = False
) -> str:
    return transfer_data_to_json_str(
        transfer_json_str_to_data(json_str=json_str, fast=fast), beautify=beautify, fast=fast
    )


def write_dicts_iterly_in_json(
        *,
        filepath: str,
        dict_iterable: Iterable[dict],
        beautify: bool = False
):
    with open(filepath, "w", encoding="utf-8") as f:
        if beautify:
            f.write("[\n")
        else:
            f.write("[")
        first_item = True

        for dict_ in dict_iterable:
            if not first_item:
                if beautify:
                    f.write(",\n")
                else:
                    f.write(",")
            if beautify:
                json.dump(dict_, f, ensure_ascii=False, indent=2)
            else:
                json.dump(dict_, f, ensure_ascii=False)
            first_item = False

        if beautify:
            f.write("\n]")
        else:
            f.write("]")

    return filepath


def json_stream_iterator(
        dict_iterable: Iterable[dict],
):
    yield "["

    first = True
    for dict_ in dict_iterable:
        if not first:
            yield ","
        else:
            first = False
        yield transfer_data_to_json_str(dict_, fast=True, beautify=False)

    yield "]"


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit

from __future__ import annotations

import subprocess

from pydantic import BaseModel


class RunCmdResHasErr(Exception):
    pass


class RunCmdRes(BaseModel):
    out: str
    err: str
    return_code: int

    @property
    def has_bad_return_code(self) -> bool:
        return self.return_code != 0

    def raise_for_bad_return_code(self):
        if self.has_bad_return_code is True:
            raise RunCmdResHasErr(f"return_code={self.return_code}, err={self.err}")
        return


def run_cmd(command: str, raise_for_bad_return_code: bool = False) -> RunCmdRes:
    subprocess_res = subprocess.run(command, shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE, check=False)

    res = RunCmdRes(
        out=subprocess_res.stdout.decode(),
        err=subprocess_res.stderr.decode(),
        return_code=subprocess_res.returncode
    )
    if raise_for_bad_return_code is True:
        res.raise_for_bad_return_code()

    return res


def __example():
    pass


if __name__ == '__main__':
    __example()



# arpakit
from arpakitlib.ar_enumeration_util import Enumeration


class BaseBlank:
    class Languages(Enumeration):
        rus = "ru"  # Русский
        eng = "en"  # Английский
        spa = "es"  # Испанский
        deu = "de"  # Немецкий
        fra = "fr"  # Французский
        ita = "it"  # Итальянский
        por = "pt"  # Португальский
        jpn = "ja"  # Японский
        kor = "ko"  # Корейский
        zho = "zh"  # Китайский
        ara = "ar"  # Арабский

    def __init__(self, *, lang: str = "ru", **kwargs):
        self.lang = lang.strip()

    def compare_lang(self, v: str) -> bool:
        if self.lang.lower() == v.lower().strip():
            return True
        return False

    def hello_world(self) -> str:
        if self.compare_lang(self.Languages.rus):
            return "Привет, мир!"  # Русский

        if self.compare_lang(self.Languages.eng):
            return "Hello, world!"  # Английский

        if self.compare_lang(self.Languages.spa):
            return "¡Hola, mundo!"  # Испанский

        if self.compare_lang(self.Languages.deu):
            return "Hallo, Welt!"  # Немецкий

        if self.compare_lang(self.Languages.fra):
            return "Bonjour, le monde!"  # Французский

        if self.compare_lang(self.Languages.ita):
            return "Ciao, mondo!"  # Итальянский

        if self.compare_lang(self.Languages.por):
            return "Olá, mundo!"  # Португальский

        if self.compare_lang(self.Languages.jpn):
            return "こんにちは、世界！"  # Японский

        if self.compare_lang(self.Languages.kor):
            return "안녕, 세상!"  # Корейский

        if self.compare_lang(self.Languages.zho):
            return "你好，世界！"  # Китайский

        if self.compare_lang(self.Languages.ara):
            return "مرحبا بالعالم!"  # Арабский

        # если язык не найден
        return "Hello, world!"



# arpakit

import pyminizip


def make_zip_with_passwd(
        *,
        input_filepath: str,
        output_filename: str = "archive.zip",
        passwd: str = "123"
) -> str:
    pyminizip.compress(
        input_filepath,  # исходный файл
        None,
        output_filename,  # архив
        passwd,  # пароль
        5  # уровень сжатия (1–9)
    )
    return output_filename

